/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/IK/DifferentialIK.h>

#include "ArmarXCore/util/CPPUtility/TripleBuffer.h"

#include <RobotAPI/components/units/RobotUnit/ControlTargets/ControlTarget1DoFActuator.h>
#include <RobotAPI/components/units/RobotUnit/SensorValues/SensorValue1DoFActuator.h>

#include <armarx/control/common/control_law/common.h>


namespace armarx::control::common::control_law
{

    class TaskspaceImpedanceController
    {
    public:
        /// you can set the following values from outside of the rt controller via Ice interfaces
        struct Config
        {
            Eigen::Vector6f kpImpedance;
            Eigen::Vector6f kdImpedance;

            Eigen::VectorXf kpNullspace;
            Eigen::VectorXf kdNullspace;

            Eigen::Matrix4f desiredPose;
            Eigen::Vector6f desiredTwist;
            std::optional<Eigen::VectorXf> desiredNullspaceJointAngles;

            float torqueLimit;
            float qvelFilter;
        };

        /// internal status of the controller, containing intermediate variables, mutable targets
        struct RtStatus
        {
            /// targets
            Eigen::VectorXf desiredJointTorques;
            Eigen::Matrix4f desiredPose;

            /// task space variables
            Eigen::Vector6f forceImpedance;
        };

        struct NonRtStatus
        {
            /// joint space variable
            Eigen::VectorXf qpos;
            Eigen::VectorXf qvel;

            /// task space variables
            Eigen::Matrix4f currentPose;
            Eigen::Vector6f currentTwist;

            Eigen::Matrix4f desiredPose;
            Eigen::Vector6f desiredTwist;

            /// others
            Eigen::MatrixXf jacobi;
            Eigen::MatrixXf jtpinv;
        };

    private:
        /// joint space variables
        int numOfJoints;
        std::atomic_bool enablePreactivateInit{false};

        Eigen::MatrixXf I;

        VirtualRobot::DifferentialIKPtr ik;

        const float lambda = 2.0f;

    public:
        TripleBuffer<RtStatus> bufferRtToOnPublish;

        TripleBuffer<NonRtStatus> bufferNonRtToRt;
        TripleBuffer<NonRtStatus> bufferNonRtToOnPublish;
        TripleBuffer<NonRtStatus> bufferNonRtToUser;

        VirtualRobot::RobotNodePtr tcp;

        void initialize(const VirtualRobot::RobotNodeSetPtr& rns);
        void preactivateInit(const VirtualRobot::RobotNodeSetPtr& rns);
        bool updateControlStatus(
                const Config& cfg,
                const RobotStatus& robotStatus);

        void run(const Config& cfg, ControlTarget& targets);
        void firstRun();

        std::atomic<bool> isInitialized{false};
    };
} /// namespace armarx::control::common::control_law
